#!/usr/bin/python3

from http.server import HTTPServer, BaseHTTPRequestHandler
from urllib.parse import urlparse, parse_qs, unquote
import os
import errno
import time
import sys
import random


from selenium import webdriver
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium.common.exceptions import TimeoutException
from selenium.webdriver.support.ui import Select
from selenium.webdriver.firefox.webdriver import FirefoxProfile
from selenium import webdriver


driverInstances = 2

drivers = []
usedDrivers = []

for driverid in range(driverInstances):
	webd = webdriver.Firefox()
	drivers.append(webd)
	webd.implicitly_wait(2)
	webd.set_page_load_timeout(9999999)
	usedDrivers.append(False)

# Port number the web server listens to
PORT_NUMBER = 23080
# Number of executed commands
commands_executed = 0





###
#name getUsedDriverInstances()
#description Gives back the lowest free webdriver instance ID or -1 if all instances are used.
#returns Lowest free webdriver instance ID or -1 if all instances are used
###
def getUsedDriverInstances():
	global usedDrivers
	global drivers

	for driverid in range(driverInstances):
		if usedDrivers[driverid] is False:
			usedDrivers[driverid] = True
			# Quit this webdriver
			drivers[driverid].quit()
			# Create a new webdriver in the same place
			drivers[driverid] = webdriver.Firefox()
			drivers[driverid].implicitly_wait(2)
			drivers[driverid].set_page_load_timeout(9999999)
			
			print('getUsedDriverInstances:' + str(driverid))
			
			return(driverid)

	return(-1)





###
#name freeDriverInstances(driverid)
#description Frees a driver instance
#parameter driverid: ID of the webdriver instance
###
def freeDriverInstances(driverid):
	global usedDrivers
	usedDrivers[driverid] = False
	print('freeDriverInstances:' + str(driverid))





###
#name wait_for_size(driver)
#description Waits until the page doesn't load more data
#parameter driver: webdriver instance
###
def wait_for_size(driver):
	olen = 1
	nlen = 2
	while nlen != olen:
		olen = len(driver.page_source)
		time.sleep(3)
		nlen = len(driver.page_source)





###
#name webwrite(self, msg)
#description Gives out a message on the generated web page.
#parameter self: Instance to the handler object
#parameter msg: Message to give out.
###
def webwrite(self, msg):
	if (type(msg) == 'byte'):
		self.wfile.write(msg)
	else:
		self.wfile.write(msg.encode())




###
#name getElem(self, driver, params)
#description Tries to get the selenium element ID for an element
#parameter self: Instance to the handler object
#parameter driver: webdriver instance
#parameter params: Dictionary with the command and its parameters
#returns: Found selenium element ID or None, if the element could not be found
###
def getElem(self, driver, params):
	if ('ID' in params):
		try:
			elem = driver.find_element_by_id(params['ID'][0])
		except:
			elem = None

	elif ('name' in params):
		try:
			elem = driver.find_element_by_name(params['name'][0])
		except:
			elem = None
	else:
		return(None)

	if (elem is None):
		raise	# So "except" of myHandler::do_GET will call a 2nd time webexec
		return(None)
	else:
		return(elem)





###
#name webexec(self, driver, params):
#description Executes the selenium commands.
#parameter self: Instance to the handler object
#parameter driver: webdriver instance
#parameter params: Dictionary with the command and its parameters
###
def webexec(self, driver, params):
	global commands_executed

	# Opens an URL in the browser window
	# parameter cmd: open
	# parameter url: urlencode URL
	if (params['cmd'][0] == "open"):
		#self.wfile.write(b'offem')
		#self.wfile.write(params['cmd'][0].encode())
		if ('url' in params):
			#print('>>>>>>>>>>>:' + params['url'][0])
			driver.get(unquote(params['url'][0]))
			#wait_for_size(driver)
			#self.wfile.write(params['url'][0].encode())



	# Closes the browser window
	# parameter cmd: close
	elif (params['cmd'][0]  == 'close'):
		driver.close()



	# Clicks a button
	# parameter cmd: clickButton
	# parameter ID: ID of the button
	# parameter name: Name of the button
	elif (params['cmd'][0]  == 'clickButton'):
		elem = getElem(self, driver, params)
		if (elem is not None):
			elem.click()
			#wait_for_size(driver)



	# Types into an editline or textarea
	# parameter cmd: typeInto
	# parameter ID: ID of the edit field
	# parameter name: Name of the edit field
	# parameter text: Urlencoded text to type into
	elif (params['cmd'][0]  == 'typeInto'):
		if ('text' in params):
			elem = getElem(self, driver, params)
			if (elem is not None):
				elem.clear()
				elem.send_keys(unquote(params['text'][0]))



	# Selects an option from a selection
	# parameter cmd: selectFrom
	# parameter ID: ID of the selection
	# parameter name: Name of the selection
	# parameter val: The value of the option to select
	elif (params['cmd'][0]  == 'selectFrom'):
		if ('val' in params):
			elem = getElem(self, driver, params)
			print(elem)
			if (elem is not None):
				try:
					print(unquote(params['val'][0]))
					Select(elem).select_by_value(unquote(params['val'][0]))
					#wait_for_size(driver)
				except:
					webwrite(self, 'ERROR:value not found')



	# Deselects an option from a selection
	# parameter cmd: deselectFrom
	# parameter ID: ID of the selection
	# parameter name: Name of the selection
	# parameter val: The value of the option to deselect
	elif (params['cmd'][0]  == 'deselectFrom'):
		if ('val' in params):
			
			elem = getElem(self, driver, params)
			if (elem is not None):
				try:
					Select(elem).deselect_by_value(unquote(params['val'][0]))
					#wait_for_size(driver)
				except:
					webwrite(self, 'ERROR:value not found')



	# Selects a radio button
	# parameter cmd: selectRadio
	# parameter ID: ID of the radio button
	# parameter name: Name of the radio button
	# parameter val: The value of the radio button to select
	elif (params['cmd'][0]  == 'selectRadio'):
		if ('val' in params):
			if ('ID' in params):
				try:
					driver.find_elements_by_xpath("//input[@id='"+params['ID'][0]+"' and @value='"+unquote(params['val'][0])+"']")[0].click()
				except:
					raise	# So "except" of myHandler::do_GET will call a 2nd time webexec
			elif ('name' in params):
				try:
					driver.find_elements_by_xpath("//input[@name='"+params['name'][0]+"' and @value='"+unquote(params['val'][0])+"']")[0].click()
				except:
					raise	# So "except" of myHandler::do_GET will call a 2nd time webexec



	# Puts the current page source of the selenium browser into the output page
	# parameter cmd: getsource
	elif (params['cmd'][0]  == 'getsource'):
		webwrite(self, driver.page_source)



	# (Un)Ticks a checkbox
	# parameter cmd: setCheck
	# parameter ID: ID of the checkbox
	# parameter name: Name of the checkbox
	# parameter checked: 1 for activating the checkbox, 0 for unticking it
	elif (params['cmd'][0]  == 'setCheck'):
		if ('checked' in params):
			checkbox =  getElem(self, driver, params)

			if (checkbox is not None):
				if (params['checked'][0] == '1'):
					checked = True
				else:
					checked = False
		
				#print(checkbox.is_selected())
				#print(checked)
		
				if (checkbox.is_selected() is not checked):
					checkbox.click()

	webwrite(self, '\n')
	commands_executed += 1


	# Closes the browser and webserver and quits the script
	# parameter cmd: quit
	if (params['cmd'][0] == 'quit'):
		driver.quit()
		server.socket.close()
		quit()


	# Reloads the loaded page
	# parameter cmd: reload
	if (params['cmd'][0] == 'reload'):
		driver.refresh()





###
#name showHelp(self)
#description Shows a help text in the terminal and as output of the GET call
#parameter self: Instance to the handler object
###
def showHelp(self):
	msg = 'Put "http://' + self.address_string() + ':' + str(PORT_NUMBER) + '''" in front of each of the following commands:

/nextdriverid: Give back the next driver ID or -1, if all are used

/free?driverid=<Driver ID>: Frees a webdriver instance

/help: Shows a help text in the terminal and as output of the GET call

/status: Shows a status text in the terminal and as output of the GET call

/run?cmd=open&url=<urlencode URL>: Opens an URL in the browser window

/run?cmd=getsource: Puts the current page source of the selenium browser into the output page

/run?cmd=close: Closes the browser window

/run?cmd=quit: Closes the browser and webserver and quits the script

/run?driverid=<Driver ID>&cmd=clickButton&name=<Name of the button>: Clicks a button
/run?driverid=<Driver ID>&cmd=clickButton&ID=<ID of the button>: Clicks a button

/run?driverid=<Driver ID>&cmd=typeInto&name=<Name of the edit field>&text=<Urlencoded text to type into>: Types into an editline or textarea
/run?driverid=<Driver ID>&cmd=typeInto&ID=<ID of the edit field>&text=<Urlencoded text to type into>: Types into an editline or textarea

/run?driverid=<Driver ID>&cmd=selectFrom&name=<Name of the selection>&val=<The value of the option to select>: Selects an option from a selection
/run?driverid=<Driver ID>&cmd=selectFrom&ID=<ID of the selection>&val=<The value of the option to select>: Selects an option from a selection

/run?driverid=<Driver ID>&cmd=deselectFrom&name=<Name of the selection>&val=<The value of the option to deselect>: Deselects an option from a selection
/run?driverid=<Driver ID>&cmd=deselectFrom&ID=<ID of the selection>&val=<The value of the option to deselect>: Deselects an option from a selection

/run?driverid=<Driver ID>&cmd=selectRadio&name=<Name of the radio button>&val=<The value of the radio button>: Selects a radio button
/run?driverid=<Driver ID>&cmd=selectRadio&ID=<ID of the radio button>&val=<The value of the radio button>: Selects a radio button

/run?driverid=<Driver ID>&cmd=setCheck&name=<Name of the checkbox>&checked=<1 for activating the checkbox, 0 for unticking it>: (Un)Ticks a checkbox
/run?driverid=<Driver ID>&cmd=setCheck&ID=<ID of the checkbox>&checked=<1 for activating the checkbox, 0 for unticking it>: (Un)Ticks a checkbox

'''

	print(msg)
	webwrite(self, msg)




###
#name showStatus(self)
#description Shows a status text in the terminal and as output of the GET call
#parameter self: Instance to the handler object
###
def showStatus(self):
	msg = 'commands_executed:' + str(commands_executed)
	showMessage(self, msg)





###
#name showMessage(self, msg)
#description Shows a message in the terminal and as output of the GET call
#parameter self: Instance to the handler object
###
def showMessage(self, msg):
	print(msg)
	webwrite(self, msg)





# Class for HTTPServer that will handle incoming GET requests
class myHandler(BaseHTTPRequestHandler):

	# Handler for the GET requests
	def do_GET(self):
		self.send_response(200)
		self.send_header('Content-type','text/html')
		self.end_headers()

		# self.path contains the full URL with all GET prameters
		urlcomp = urlparse(self.path)

		# urlcomp.query contains the GET parameters string
		params = parse_qs(urlcomp.query)
		# params contains a dictionary with the GET keys as keys and the GET values as *list* (use [0] to get the actual value)
		
		# Check, if the driverid is given in the URL
		if ('driverid' in params):
			driverid = int(params['driverid'][0])
			driver = drivers[driverid]
		else:
			driverid = False
			driver = False
		
		# urlcomp.path contains the name of the requested file with full path
		if (urlcomp.path == '/run'):
			# urlcomp.query contains the GET parameters string
			params = parse_qs(urlcomp.query)
			# params contains a dictionary with the GET keys as keys and the GET values as *list* (use [0] to get the actual value)
			
			# Check, if the driverid is given in the URL
			if (driverid is not False):
				# The selenium commands have to be run twice most time, due to "broken pipe" error on the first run
				try:
					webexec(self, driver, params)
				except:
					webexec(self, driver, params)
			else:
				showMessage(self, 'No driverid given in the URL')

		if (urlcomp.path  == '/help'):
			showHelp(self)

		if (urlcomp.path  == '/free'):
			showMessage(self, 'Freed webdriver instance #' + str(driverid))
			freeDriverInstances(driverid)

		if (urlcomp.path  == '/status'):
			showStatus(self)

		# Give back the next driver id or -1, if all are used
		if (urlcomp.path  == '/nextdriverid'):
			showMessage(self, str(getUsedDriverInstances()))




try:
	# Start the webserver on the given port with custom GET handler
	server = HTTPServer(('', PORT_NUMBER), myHandler)
	print('HTTP to Selenium bridge listening on port ' , PORT_NUMBER)
	server.serve_forever()


except KeyboardInterrupt:
	print('^C received, shutting down the web server')
server.socket.close()
